; SPEEDF84.ASM - Morse Code Speedometer using the 16F84 PIC
;
;
; 1/31/2000     Chuck Olson, WB9KZY
;               Jackson Harbor Press
;               http://jacksonharbor.home.att.net
;               jacksonharbor@att.net
;
;
;       What does it do?
;
;       This program uses a 16f84 to measure the "dit speed" of a Morse
;       code stream and then display the speed in wpm from 1 to 99 wpm
;       via either a 2 digit LED display or sent in Morse code to a piezo 
;       speaker.
;
;       See the file: SPEEDREA.DME for more information on the program.
;
;
;       Possible Enhancements:
;
;       0) eliminate the hc595 shift registers for the LED display by using
;               PORTB for a multiplexed drive output of the 2 LEDs.
;       1) have an optional character or word speed mode, to more accurately
;               judge the speed when someone is "Farnsworthing".
;       2) add an LCD output on PORTB
;       3) change the measurement from the dit interval only, to the dit
;               interval PLUS the dit space interval - also measure the
;               dah interval in the same fashion - needed for "weighted"
;               sending, especially for manual sending.
;       4) increase the resolution of the measurement which would allow
;               more accurate readings, especially at the top end. 
;               possiby add a third digit of output - to a tenth of a wpm
;       5) add a decimal diagnostic mode to allow outputting the actual
;               interval length to the piezo in decimal rather than hex.
;       6) add a sleep mode (more for RFI quiet rather than for power
;               consumption)
;       7) add a display timeout to turn off the LEDs after a certain
;               period of inactivity
;       8) add some type of averaging or median filtering to prevent 
;               display "jumpiness"
;       9) add a weight display - send a 2 digit percentage - the percentage
;               is the duty cycle of the weighted dit - an ideal 1:3 dit
;               to dah ratio will have a weight percentage of 50% (the dit
;               interval and the dit space interval are the same)
;       
;
;       PIC pinout - see the file:  SPEED.GIF for the project schematic
;
;PIC
;pin    name    function/connection
;---    ----    -----------------------------------------------
;  1    ra2     key input - senses a dit/dah as logic 0, space as logic 1
;  2    ra3     select - pin 12 of the hc595
;  3    ra4     strobe - pin 11 of the hc595
;  4    MCLR    pulled up to +5V with a 10K resistor
;  5    VSS     Ground 
;  6    rb0     
;  7    rb1     no connect
;  8    rb2     no connect
;  9    rb3     no connect
; 10    rb4     
; 11    rb5     
; 12    rb6     
; 13    rb7     
; 14    VDD     +5V
; 15    osc2    connected to 4 MHz xtal & 22? pf cap
; 16    osc1    connected to 4 MHz xtal & 22? pf cap
; 17    ra0     datao - pin 14 of the hc595
; 18    ra1     piezo speaker output for sending the speed in Morse


	LIST P=16F84, R=DEC 
	INCLUDE "p16F84.inc"
	__FUSES _PWRTE_OFF & _CP_OFF & _WDT_OFF & _XT_OSC
; powerup timer off - not sure why.
; copy protect is off
; watchdog timer is off 
; the oscillator uses an external 4 Mhx crystal or resonator


; memory locations
tens    equ     H'0C'   ;holder for tens digit of converted speed byte
ones    equ     H'0D'   ;holder for ones digit of converted speed byte
temp    equ     H'0E'   ;all purpose temporary byte
highcb  equ     H'0F'   ;high key down counter byte
lowcb   equ     H'10'   ;low key down counter byte
char    equ     H'11'   ;holder for character data - used in playchr, strtbit
size    equ     H'12'   ;char size counter - used in playchr, strtbit
third   equ     H'13'   ;holder for number of dit length spaces => ditsp
temp2   equ     H'14'   ;holder for element length => outdit, outdah, ditsp
wpm     equ     H'15'   ;wpm counter - pointer to the length table => testlen
lastdhi equ     H'16'   ;last dit interval high byte
lastdlo equ     H'17'   ;last dit interval low byte
temp3   equ     H'18'   ;holder for element length / 2 - high byte
temp4   equ     H'19'   ;holder for element length / 2 - low byte


; constants - also see individual routines


; bit definitions
in      equ     2       ;bit 2 of port a is the input pin
carry   equ     0       ;bit 0 of STATUS is the carry bit
select  equ     3       ;select is pin 12 of the hc595 - bit 3 of PORTA
strobe  equ     4       ;strobe is pin 11 of the hc595 - bit 4 of PORTA
datao   equ     0       ;datao is pin 14 of the hc595 - bit 0 of PORTA
zero    equ     2       ;bit 2 of STATUS is the zero bit
switch  equ     0       ;bit 0 of PORTB is the switch input pin


	ORG     0x00
start   goto    init    ;startup routine
	

	ORG     0x04
rupt    goto    init    ;interrupt routine


; table of equivalent morse characters
;character structure from Dave Gwillym, KB2TQX
; right justified storage of the character:
;                                       dit = 0
;                                       dah = 1
; the leftmost 1 is a start bit - the byte is left shifted until a
;       1 is seen, then the dits and dahs are shifted out for a
;       total of 8 shifts - this holds most normal characters.
; note that the table position is what is stored in the memory
;
numbers addwf   PCL,1   ;compute the jump value
;                             character  hex      table position        
	retlw   b'00111111'   ;0 ascii   30 hex   00
	retlw   b'00101111'   ;1         31       01
	retlw   b'00100111'   ;2         32       02
	retlw   b'00100011'   ;3         33       03
	retlw   b'00100001'   ;4         34       04
	retlw   b'00100000'   ;5         35       05
	retlw   b'00110000'   ;6         36       06
	retlw   b'00111000'   ;7         37       07
	retlw   b'00111100'   ;8         38       08
	retlw   b'00111110'   ;9         39       09
;1/27/00 - eliminated the hex values below - they are only needed for
;       sending the raw data to the piezo in hex (sendit, maskns routines)
;lets    retlw   b'00000101'   ;A         3A hex   0A
;        retlw   b'00011000'   ;B         3B       0B
;        retlw   b'00011010'   ;C         3C       0C
;        retlw   b'00001100'   ;D         3D       0D
;        retlw   b'00000010'   ;E         3E       0E
;        retlw   b'00010010'   ;F         3F       0F


; length - a 2 byte wide table of dit interval decision points.  It
;       was generated with the Qbasic program, speedtab.bas - then the 
;       table was edited to get rid of the spaces around the decimal value.
;       The decision points are at half word per minute (wpm) intervals.
;       If the measured dit interval is greater than the present decision 
;       point value and less than the previous value, the speed is simply
;       the table pointer divided by 2 (for the table width).  The decision
;       points are presented with the most significan byte first.
length  addwf   PCL,1   ;compute the jump value
;        MSB then LSB    Speed      dit duration in .1 millisecond chunks
	retlw   d'93'  ; .5 wpm     24000 x .1 ms
	retlw   d'192'
	retlw   d'31'  ; 1.5 wpm     8000 x .1 ms
	retlw   d'64'
	retlw   d'18'  ; 2.5 wpm     4800 x .1 ms
	retlw   d'192'
	retlw   d'13'  ; 3.5 wpm     3429 x .1 ms
	retlw   d'101'
	retlw   d'10'  ; 4.5 wpm     2667 x .1 ms
	retlw   d'107'
	retlw   d'8'  ; 5.5 wpm     2182 x .1 ms
	retlw   d'134'
	retlw   d'7'  ; 6.5 wpm     1846 x .1 ms
	retlw   d'54'
	retlw   d'6'  ; 7.5 wpm     1600 x .1 ms
	retlw   d'64'
	retlw   d'5'  ; 8.5 wpm     1412 x .1 ms
	retlw   d'132'
	retlw   d'4'  ; 9.5 wpm     1263 x .1 ms
	retlw   d'239'
	retlw   d'4'  ; 10.5 wpm     1143 x .1 ms
	retlw   d'119'
	retlw   d'4'  ; 11.5 wpm     1043 x .1 ms
	retlw   d'19'
	retlw   d'3'  ; 12.5 wpm     960 x .1 ms
	retlw   d'192'
	retlw   d'3'  ; 13.5 wpm     889 x .1 ms
	retlw   d'121'
	retlw   d'3'  ; 14.5 wpm     828 x .1 ms
	retlw   d'60'
	retlw   d'3'  ; 15.5 wpm     774 x .1 ms
	retlw   d'6'
	retlw   d'2'  ; 16.5 wpm     727 x .1 ms
	retlw   d'215'
	retlw   d'2'  ; 17.5 wpm     686 x .1 ms
	retlw   d'174'
	retlw   d'2'  ; 18.5 wpm     649 x .1 ms
	retlw   d'137'
	retlw   d'2'  ; 19.5 wpm     615 x .1 ms
	retlw   d'103'
	retlw   d'2'  ; 20.5 wpm     585 x .1 ms
	retlw   d'73'
	retlw   d'2'  ; 21.5 wpm     558 x .1 ms
	retlw   d'46'
	retlw   d'2'  ; 22.5 wpm     533 x .1 ms
	retlw   d'21'
	retlw   d'1'  ; 23.5 wpm     511 x .1 ms
	retlw   d'255'
	retlw   d'1'  ; 24.5 wpm     490 x .1 ms
	retlw   d'234'
	retlw   d'1'  ; 25.5 wpm     471 x .1 ms
	retlw   d'215'
	retlw   d'1'  ; 26.5 wpm     453 x .1 ms
	retlw   d'197'
	retlw   d'1'  ; 27.5 wpm     436 x .1 ms
	retlw   d'180'
	retlw   d'1'  ; 28.5 wpm     421 x .1 ms
	retlw   d'165'
	retlw   d'1'  ; 29.5 wpm     407 x .1 ms
	retlw   d'151'
	retlw   d'1'  ; 30.5 wpm     393 x .1 ms
	retlw   d'137'
	retlw   d'1'  ; 31.5 wpm     381 x .1 ms
	retlw   d'125'
	retlw   d'1'  ; 32.5 wpm     369 x .1 ms
	retlw   d'113'
	retlw   d'1'  ; 33.5 wpm     358 x .1 ms
	retlw   d'102'
	retlw   d'1'  ; 34.5 wpm     348 x .1 ms
	retlw   d'92'
	retlw   d'1'  ; 35.5 wpm     338 x .1 ms
	retlw   d'82'
	retlw   d'1'  ; 36.5 wpm     329 x .1 ms
	retlw   d'73'
	retlw   d'1'  ; 37.5 wpm     320 x .1 ms
	retlw   d'64'
	retlw   d'1'  ; 38.5 wpm     312 x .1 ms
	retlw   d'56'
	retlw   d'1'  ; 39.5 wpm     304 x .1 ms
	retlw   d'48'
	retlw   d'1'  ; 40.5 wpm     296 x .1 ms
	retlw   d'40'
	retlw   d'1'  ; 41.5 wpm     289 x .1 ms
	retlw   d'33'
	retlw   d'1'  ; 42.5 wpm     282 x .1 ms
	retlw   d'26'
	retlw   d'1'  ; 43.5 wpm     276 x .1 ms
	retlw   d'20'
	retlw   d'1'  ; 44.5 wpm     270 x .1 ms
	retlw   d'14'
	retlw   d'1'  ; 45.5 wpm     264 x .1 ms
	retlw   d'8'
	retlw   d'1'  ; 46.5 wpm     258 x .1 ms
	retlw   d'2'
;
; Note: The higher order part of the table isn't used in the program (due 
;       to the debounce routine), it has been commented for reference.  
;       To actually use a speedometer at the highest speeds,
;       it would probably be a good idea to increase the resolution of
;       the measure routine to .01 ms (10 us) to give greater separation
;       between the wpm entries - as it is, most of the higher wpm values
;       only differ by .1 ms or so.
;
;        retlw   d'0'  ; 47.5 wpm     253 x .1 ms
;        retlw   d'253'
;        retlw   d'0'  ; 48.5 wpm     247 x .1 ms
;        retlw   d'247'
;        retlw   d'0'  ; 49.5 wpm     242 x .1 ms
;        retlw   d'242'
;        retlw   d'0'  ; 50.5 wpm     238 x .1 ms
;        retlw   d'238'
;        retlw   d'0'  ; 51.5 wpm     233 x .1 ms
;        retlw   d'233'
;        retlw   d'0'  ; 52.5 wpm     229 x .1 ms
;        retlw   d'229'
;        retlw   d'0'  ; 53.5 wpm     224 x .1 ms
;        retlw   d'224'
;        retlw   d'0'  ; 54.5 wpm     220 x .1 ms
;        retlw   d'220'
;        retlw   d'0'  ; 55.5 wpm     216 x .1 ms
;        retlw   d'216'
;        retlw   d'0'  ; 56.5 wpm     212 x .1 ms
;        retlw   d'212'
;        retlw   d'0'  ; 57.5 wpm     209 x .1 ms
;        retlw   d'209'
;        retlw   d'0'  ; 58.5 wpm     205 x .1 ms
;        retlw   d'205'
;        retlw   d'0'  ; 59.5 wpm     202 x .1 ms
;        retlw   d'202'
;        retlw   d'0'  ; 60.5 wpm     198 x .1 ms
;        retlw   d'198'
;        retlw   d'0'  ; 61.5 wpm     195 x .1 ms
;        retlw   d'195'
;        retlw   d'0'  ; 62.5 wpm     192 x .1 ms
;        retlw   d'192'
;        retlw   d'0'  ; 63.5 wpm     189 x .1 ms
;        retlw   d'189'
;        retlw   d'0'  ; 64.5 wpm     186 x .1 ms
;        retlw   d'186'
;        retlw   d'0'  ; 65.5 wpm     183 x .1 ms
;        retlw   d'183'
;        retlw   d'0'  ; 66.5 wpm     180 x .1 ms
;        retlw   d'180'
;        retlw   d'0'  ; 67.5 wpm     178 x .1 ms
;        retlw   d'178'
;        retlw   d'0'  ; 68.5 wpm     175 x .1 ms
;        retlw   d'175'
;        retlw   d'0'  ; 69.5 wpm     173 x .1 ms
;        retlw   d'173'
;        retlw   d'0'  ; 70.5 wpm     170 x .1 ms
;        retlw   d'170'
;        retlw   d'0'  ; 71.5 wpm     168 x .1 ms
;        retlw   d'168'
;        retlw   d'0'  ; 72.5 wpm     166 x .1 ms
;        retlw   d'166'
;        retlw   d'0'  ; 73.5 wpm     163 x .1 ms
;        retlw   d'163'
;        retlw   d'0'  ; 74.5 wpm     161 x .1 ms
;        retlw   d'161'
;        retlw   d'0'  ; 75.5 wpm     159 x .1 ms
;        retlw   d'159'
;        retlw   d'0'  ; 76.5 wpm     157 x .1 ms
;        retlw   d'157'
;        retlw   d'0'  ; 77.5 wpm     155 x .1 ms
;        retlw   d'155'
;        retlw   d'0'  ; 78.5 wpm     153 x .1 ms
;        retlw   d'153'
;        retlw   d'0'  ; 79.5 wpm     151 x .1 ms
;        retlw   d'151'
;        retlw   d'0'  ; 80.5 wpm     149 x .1 ms
;        retlw   d'149'
;        retlw   d'0'  ; 81.5 wpm     147 x .1 ms
;        retlw   d'147'
;        retlw   d'0'  ; 82.5 wpm     145 x .1 ms
;        retlw   d'145'
;        retlw   d'0'  ; 83.5 wpm     144 x .1 ms
;        retlw   d'144'
;        retlw   d'0'  ; 84.5 wpm     142 x .1 ms
;        retlw   d'142'
;        retlw   d'0'  ; 85.5 wpm     140 x .1 ms
;        retlw   d'140'
;        retlw   d'0'  ; 86.5 wpm     139 x .1 ms
;        retlw   d'139'
;        retlw   d'0'  ; 87.5 wpm     137 x .1 ms
;        retlw   d'137'
;        retlw   d'0'  ; 88.5 wpm     136 x .1 ms
;        retlw   d'136'
;        retlw   d'0'  ; 89.5 wpm     134 x .1 ms
;        retlw   d'134'
;        retlw   d'0'  ; 90.5 wpm     133 x .1 ms
;        retlw   d'133'
;        retlw   d'0'  ; 91.5 wpm     131 x .1 ms
;        retlw   d'131'
;        retlw   d'0'  ; 92.5 wpm     130 x .1 ms
;        retlw   d'130'
;        retlw   d'0'  ; 93.5 wpm     128 x .1 ms
;        retlw   d'128'
;        retlw   d'0'  ; 94.5 wpm     127 x .1 ms
;        retlw   d'127'
;        retlw   d'0'  ; 95.5 wpm     126 x .1 ms
;        retlw   d'126'
;        retlw   d'0'  ; 96.5 wpm     124 x .1 ms
;        retlw   d'124'
;        retlw   d'0'  ; 97.5 wpm     123 x .1 ms
;        retlw   d'123'
;        retlw   d'0'  ; 98.5 wpm     122 x .1 ms
;        retlw   d'122'
;        retlw   d'0'  ; 99.5 wpm     121 x .1 ms
;        retlw   d'121'


; sevseg - 10 bytes that define the segment data for 0-9 on the LED
;
;  bit  7  6  5  4  3  2  1  0
;  seg  a  b  c  d  e  f  g  -
;
;      +----a----+
;      |         |
;      f         b          1 = on
;      |         |          0 = off
;      +----g----+          common cathode
;      |         |
;      e         c
;      |         |
;      +----d----+  
sevseg  addwf   PCL,1   ;compute the jump value
;                 abcdefg
	retlw   b'11111100'    ;0
	retlw   b'01100000'    ;1
	retlw   b'11011010'    ;2
	retlw   b'11110010'    ;3
	retlw   b'01100110'    ;4
	retlw   b'10110110'    ;5
	retlw   b'10111110'    ;6
	retlw   b'11100000'    ;7
	retlw   b'11111110'    ;8
	retlw   b'11110110'    ;9
; hex display (A-F) not needed in this program
;        retlw   b'11101110'    ;A
;        retlw   b'00111110'    ;b
;        retlw   b'10011100'    ;C
;        retlw   b'01111010'    ;d
;        retlw   b'10011110'    ;E
;        retlw   b'10001110'    ;F
			  

; 1/28/00 commented out the following common anode version of the
;       seven segment table
;; sevseg - 10 bytes that define the segment data for 0-9 on the LED
;;
;;  bit  7  6  5  4  3  2  1  0
;;  seg  a  b  c  d  e  f  g  -
;;
;;      +----a----+
;;      |         |
;;      f         b          1 = on
;;      |         |          0 = off
;;      +----g----+          common anode
;;      |         |
;;      e         c
;;      |         |
;;      +----d----+  
;sevseg  addwf   PCL,1   ;compute the jump value
;;                 abcdefg
;        retlw   b'00000010'    ;0
;        retlw   b'10011110'    ;1
;        retlw   b'00100100'    ;2
;        retlw   b'00001100'    ;3
;        retlw   b'10011000'    ;4
;        retlw   b'01001000'    ;5
;        retlw   b'01000000'    ;6
;        retlw   b'00011110'    ;7
;        retlw   b'00000000'    ;8
;        retlw   b'00001000'    ;9
;; hex display (A-F) not needed in this program
;;        retlw   b'00010000'    ;A
;;        retlw   b'11000000'    ;b
;;        retlw   b'01100010'    ;C
;;        retlw   b'10000100'    ;d
;;        retlw   b'01100000'    ;E
;;        retlw   b'01110000'    ;F
			   

; outdit - a routine to send a dit
ditlen  equ     d'80'   ;(80 x 1 ms = 80 ms)
outdit  movlw   ditlen  ;load up a 15 wpm dit length (1.2 / 15 = .08)
	movwf   temp2   ;store it in the second temp location
	goto    loop4   ;enter the dah routine


; outdah - a routine to send a dah
dahlen  equ     d'240'  ;(240 x 1 ms = 240 ms)
outdah  movlw   dahlen  ;load up a 15 wpm dah length (3 x 1.2 / 15 = .24)
	movwf   temp2   ;store it in the second temp location
loop4   decfsz  temp2,1 ;decrement the dit length counter - is it 0 ?
	goto    beep4   ;no, start beeping
	;yes, end the dit
	goto    ditsp   ;with a dit space
beep4   bsf     PORTA,1 ; set piezo high
	call    halfcyc ;wait a half audio cycle
	bcf     PORTA,1 ; set piezo low
	call    halfcyc ;wait a half audio cycle
	goto    loop4   ;loop back until this dit length is complete


; ditsp - a routine to wait 1 dit space
; wdsp - a routine to wait 6 dit spaces + 1 from last dit or dah = 7 total 
; wdsp2 - a routine to wait 4 dit spaces + 3 from last char = 7 total 
; chsp - a routine to wait 2 dit spaces + 1 from last dit or dah = 3 total
ditsp   movlw   0x2     ;load up 1 dit space (add one due to loop contstruct)
	goto    loadet  ;branch in
wdsp    movlw   0x7     ;load up 6 dit spaces
	goto    loadet  ;branch in
wdsp2   movlw   0x5     ;load up 4 dit space (an additional delay after chsp)
	goto    loadet  ;branch into the routine with an alterante
chsp    movlw   0x3     ;load up 2 dit spaces
loadet  movwf   third   ;save it in third - a temp reg
decet   decfsz  third,1 ;decrement the dit space counter - is it 0 ?
	goto    ditsp2  ;no, start waiting
	goto    dun2    ;yes, end the dit
ditsp2  movlw   d'80'   ;load a 15 wpm dit length
	movwf   temp2   ;store it in the second temp location
loop6   decfsz  temp2,1 ;decrement the dit length counter - is it 0 ?
	goto    nobeep2 ;no, start waiting
	goto    decet   ;yes, check the dit space counter
nobeep2 call    halfcyc ;wait a half audio cycle
	call    halfcyc ;wait a half audio cycle
	goto    loop6   ;go back for another cycle?
dun2    return          ;end of a bunch of routines


; halfcyc - a routine to wait 1/2 of a sidetone cycle - in this case
;       that is 500 usec or 1/2 ms => 1000 Hz tone
;       at PIC clock = 4 MHz, 500 usec = 500 PIC instruction cycles
;       delay = 2 (call) + 6 (routine) + ((freq-1) x 3) loop delay
freq    equ     d'165'  ; 8 + (164 x 3) = 500
halfcyc movlw   freq    ;1   1 load up the 1/2 cycle delay constant
	movwf   temp    ;1   2 store it in the temporary location
loopy   decfsz  temp,1  ;1     decrement the temp location - is it 0 ?
	goto    loopy   ;2     no, keep decrementing
	;yes, return to the calling routine
	return          ;2 500 end of half cycle routine


; strtbit - a routine to rotate a char left until a 1 appears
strtbit movlw   0x8     ;load up the maximum char size + 1
	movwf   size    ;save it temp
strbit2 decf    size,1  ;decrement the element counter
rotit   rlf     char,1  ;rotate the msb into carry
	btfss   STATUS,0 ;test the carry - 0=placeholder, 1=start bit
	goto    strbit2 ;it's a just a placeholder (0), keep shifting
	return          ;you're done - end of the startbit routine            

	
; playchr - a routine to play a character stored in w with Morse code
playchr call    numbers ;lookup the value
	movwf   char    ;save it in char
getdig3 call    strtbit ;rotate char left until the start bit comes up
shout3  rlf     char,1  ;rotate the msb into carry
	btfsc   STATUS,0 ;test the carry - 0=dit, 1=dah
	goto    dahsnd3 ;it's a one, send a dah
	; send a dit
	call    outdit  ;send a dit
	goto    sizdec  ;go check the size variable
dahsnd3 ;btfss   ditdah,command  ;see if you should key the xmtr
	;no, just start the dah
	call    outdah  ;send the dah
sizdec  decfsz  size,1  ;decrement the char size counter, is it 0?        
	goto    shout3  ;no, go back for the next bit
	;yes, send a char space and then kill this thing
	call    chsp    ;send a character space
rpcend  return          ;end of playchr routine


; init - the initialization power up entry point of the program
init    bsf     STATUS,5        ;set the bank bit to 1
	movlw   b'00000001'     ; all outs on PORTB xcept bit 0 (switch)
	movwf   TRISB           ; set the tri-state register
	movlw   b'00000100'     ; all outputs on PORTA, except bit 2
	movwf   TRISA           ; set the tri-state register
	movlw   b'00000000'     ; no option bits are relevant except PORTB p/u
	movwf   OPTION_REG      ; set the option register
	bcf     STATUS,5        ;set the bank bit to 0
	bcf     INTCON,GIE      ;disable interrupts
	call    blanko          ;initialize the LED display      
	call    blanko          ;   with 2 blank digits
	movlw   d'10'           ;10 represents a dit longer than a 5 wpm dit
	movwf   lastdhi         ;initialize the last dit interval high < 5 wpm
	clrf    lastdlo         ; clear the low byte


mainlup ;check for input activity and branch if something happens
	btfss   PORTB,switch    ;has the Morse output switch been pressed?
	goto    morsout         ;yes, send the speed in Morse
	btfsc   PORTA,in        ;no, check the input pin                
	goto    mainlup         ;no activity, keep looping


; timeit is a routine to measure the input, key down time
;       use 100 us chunks to 
timeit  clrf    highcb          ;1 clear the high counter byte
	clrf    lowcb           ;1       and the low counter byte, too
lstart  movlw   30              ;1   1  load up a loop count
	movwf   temp            ;1   2  save it 
	btfsc   PORTA,in        ;2   4 has the key been released?
	goto    bailo           ;      yes, bail out of the loop
	;no wait for a total of 100 us including incrementing the counters
	;cdown delay is temp-1 x 3 + 2 = loop delay
	;               (30-1) x 3 + 2 = 89
cdown   decfsz  temp,1          ;1 (2)  decrement the loop count - skip at 0
	goto    cdown           ;2   93 loop back
	nop                     ;1   94 waste a cycle
	incfsz  lowcb,1         ;1   95  increment the low counter byte
	goto    evenit          ;2   97  no rollover, even out the path delay
	incfsz  highcb,1        ;1   97  increment the high counter byte
	goto    evenit2         ;2   98    no rollover, even out the path delay
nogood  clrf    highcb  ;the high order byte shouldn't rollovr
	clrf    lowcb
	goto    mainlup         ;bail back to the main loop.
evenit  nop                     ;1  98evens out the incfsz not needed for high byte
evenit2 goto    lstart          ;2 100loop back for another 100 us
bailo   ;looping is over, now find the speed measured by the dit period


; testlen - a routine to compare the measured dit length against the length
;       table of half wpm values - then adjust the table position (of the
;       first table value to be less than the measured value) into the 
;       wpm value.  This is then converted to a 2 digit decimal and then
;       sent to the piezo.
; 1/28/00 - changed testlen to use lastdhi and lastdlo which are the last dit
;       length bytes - first the measured interval is divided by 2 and then
;       compared with the last dit intervals - if the measured interval
;       reading/2 is less than the last dit, assume it's also a dit, store it 
;       as the last dit interval and display the result - if the latest 
;       reading/2 is greater than the last dit, assume that it's a dah 
;       (3 x dit / 2 = 1.5 x dit) and don't display the reading.  This 
;       technique may be fooled by weighting or by a quickly changing speed 
;       (via pot turn?).
testlen ;inserted the following 3 statements to reject any very short 
	;       intervals from straight key input - these very short intervals
	;       cause the testlen routine to hang - this is really just a 
	;       bandaid fix, the bug causing the hang should be found and 
	;       removed - however since 45 wpm is the top speed I'm interested
	;       in, I've left it like this for now.
	movf    highcb,0        ;load up the high order byte of the int count
	btfsc   STATUS,zero     ;is it zero?
	goto    mainlup         ;yes, the interval is too short, < 25.5 ms 
	bcf     STATUS,carry    ;clear the carry bit to prepare for next line
	rrf     highcb,0   ;divide the high interval byte by 2 and put into w
	movwf   temp3   ;store it temp in temp3
	rrf     lowcb,0 ;divide the low interval byte by 2 and put into w
	movwf   temp4   ;store it temp in temp4
	movf    temp3,0 ;reload the measured interval / 2
	subwf   lastdhi,0  ;subtract it from the last dit hi order byte
	btfss   STATUS,carry ;is the result negative (carry=0)?
     ;yes, this means the measured interval is a dah so ignore it
	goto    mainlup ;go back for another measurement
     ;no, see if it's equal or less than measured       
seqlt2  btfss   STATUS,2 ; a carry occurred, are the values equal? (Z=1)
	goto    adit    ;no, the measured interval is a dit
higheq2 ;yes, the high order bytes are equal, check the low order bytes
	movf    temp4,0 ;reload the low interval byte / 2
	;subwf   lowcb,0 ;subtract it from the measured lo order length byte
	subwf   lastdlo,0 ;subtract it from the last dit lo order length byte
	btfss   STATUS,carry ;is the result negative (carry=0)?
	;yes, this means the measured interval is a dah so ignore it
	goto    mainlup ;go back to the main loop
	;no, this means that the measure interval is a dit
adit    ;it's a dit, change the last dit values to the present measured values
	movf    highcb,0        ;load up the high order measured byte
	movwf   lastdhi ;save it in the high order last dit byte
	movf    lowcb,0 ;load up the low order measured byte
	movwf   lastdlo ;save it in the low order last dit byte
; next, find the wpm equivalent of the measured interval        
	clrf    wpm     ;clear the table pointer
getdat  ;first, test wpm to see if it's less than 200
	;movlw   d'200'  ;load up 200 (100 wpm x 2)
	movlw   d'92'  ;load up 92 (46 wpm x 2)
	subwf   wpm,0   ;is wpm less than 200?
	btfsc   STATUS,carry  ;is the result negative (carry = 0)?
	goto    mainlup ;no, bail out, you are too high
	;yes, you are still below 200, keep going
	movf    wpm,0   ;load the wpm value into w
	call    length  ;get the high order byte of the length table value
	subwf   highcb,0  ;subtract it from the measured hi order length byte
	btfsc   STATUS,0 ;is the result negative (carry=0)?
	goto    seqlt   ;no, see if it's equal or less than measured
	incf    wpm,1   ;yes, increment the counter twice (no need to compare
	incf    wpm,1   ;        the lower byte)
	goto    getdat  ;go back for another compare
seqlt   btfss   STATUS,2 ; a carry occurred, are the values equal? (Z=1)
	goto    display  ;no, enter the display routine
higheq  ;yes, the high order bytes are equal, check the low order bytes
	incf    wpm,1   ;increment the pointer to get the low byte
	movf    wpm,0   ;load the pointer into w
	call    length  ;       get the low order length table byte
	subwf   lowcb,0 ;subtract it from the measured lo order length byte
	btfsc   STATUS,0 ;is the result negative (carry=0)?
	goto    display ;it's equal or less than measured, display it
	incf    wpm,1   ;yes, increment the table pointer and
	goto    getdat  ; then go back for another compare


; display - a routine to send the 2 (or 1) digit wpm value to the 
;       piezo output in Morse code at 15 wpm - changed display to
;       send the data to a pair of 7 segment LEDs, serially, LSD first
display bcf     STATUS,0        ;clear the carry bit
	rrf     wpm,0   ;the table is less than the measure value, 
			;adjust wpm for display and load into w
	call    bin2dec ;convert the binary number to 2 dec digits
	;first do the 1's digit
wuns    movf    ones,0  ;get the character
	call    out     ;send the digit to the LED
overtst movf    tens,1  ;test the tens digit
	btfsc   STATUS,2 ;skip playing the char if it is zero
	goto    blank   ;it is zero, send a blank digit to the LED
	movf    tens,0  ;it's not zero, get the digit
tensout call    out     ;send the digit 
	goto    mainlup ;return to the main loop
blank   movlw   b'00000000'     ;all zeroes is a blank display digit
	call    blanko  ;send the blank digit
	goto    mainlup ;return to the main loop

	
; morsout - a routine to send the last measured speed to the 
;       piezo in Morse code (at 15 WPM)
morsout movf    tens,0  ;load the tens digit
	btfsc   STATUS,zero ;is the leading digit a 0?
	goto    vvuns   ;yes, don't send it, just proceed to the ones digit
	call    playchr ;play the character
vvuns   movf    ones,0  ;get the ones digit
	call    playchr ;play the character
	goto    mainlup ;return to the main loop


;sendit and maskns are unused diag routines to check the raw data accuracy
;; sendit - try sending the raw data to the piezo
;sendit  swapf   highcb,0        ;load up the high order byte & swap nibbles
;        call    maskns          ;send the upper nibble of the high order byte
;        movf    highcb,0        ;load up the high order byte
;        call    maskns          ;send the lower nibble of the hi order byte
;        swapf   lowcb,0         ;load up the lo order byte and swap nibbles
;        call    maskns          ;send the upper nibble of the lo order byte
;        movf    lowcb,0         ;load up the low order byte
;        call    maskns          ;send the lower nibble of the lo order byte
;        goto    mainlup         ;re-enter the main loop
;
;
;;maskns - mask and send the 4 lower bits of w to the piezo in Morse
;maskns  andlw   b'00001111'     ;mask out the upper 4 bits
;        call    numbers         ;get the Morse equivalent of the hex number
;        call    playchr         ;play this char
;        return                  ;end of maskns


; bin2dec - a routine to convert a single binary byte into a 2 decimal digit 
;       equivalent.  The routine subtracts 10 from the binary byte until
;       a negative result occurs, keeping a count of the number of times 
;       that 10 is subtracted - then the last 10 is added back - the
;       counter for the 10s subtraction contains the tens decimal digit
;       equivalent of the binary number, the holder for the binary number
;       contains the ones decimal digit equivalent.
;   memory used (destroyed): tens, ones
;   bin2dec is entered with the binary number in W.
bin2dec clrf    tens    ;clear the tens holder
	movwf   ones    ;store the binary byte into ones
incre   incf    tens,1  ;bump the tens holder
	movlw   0x0A    ;load up a 10 into w
	subwf   ones,1  ;try subtracting 10 from the binary value
	btfsc   STATUS,0 ;keep going until result is negative (carry=0)
	goto    incre   ;it's positive, keep looping
	decf    tens,1  ;too far, give back the last increment
	movlw   0x0A    ;load up a 10 into w
	addwf   ones,1  ;"    "    add back the last subtraction
	;so now the 10's digit is in tens
	; and the 1's digit is in ones
	return          ;end of bin2dec rouitne


; out - a routine to  convert the contents of w to segment data and then
;       send w bits out serially, SPI style to the HC595
blanko  movlw   b'00000000'     ;load up a blank digit
	goto    out2            ;display the blank
out     call    sevseg  ;convert the digit data in w to segment data
out2    bcf     PORTA,select    ;select the HC595s for data transfer
	movwf   temp    ;store the segment data temporarily
	movlw   8       ;setup for an 8 bit xfer
	movwf   temp2   ;store the count temp
	bcf     STATUS,carry    ;clear the carry bit
mor8    rlf     temp,1  ; rotate the segment data to shift a bit out
	btfss   STATUS,carry    ;is the bit a one or a zero
	goto    zerosnd ;send a zero
one     bsf     PORTA,datao     ;send a one
strb    bsf     PORTA,strobe    ;send strobe
	bcf     PORTA,strobe    ;clear the strobe
	decfsz  temp2,1         ;decrement the bit counter
	goto    mor8            ;it's not zero, keep going
	bsf     PORTA,select    ;deselect the 74HC595
	return                  ;end of the out routine
zerosnd bcf     PORTA,datao     ;send a zero
	goto    strb            ;strobe it in


	END     ;end of the speedf84.asm program


